import abc
import re

from metar_taf_parser.commons.converter import convert_temperature_remarks, convert_precipitation_amount
from metar_taf_parser.commons.exception import TranslationError
from metar_taf_parser.commons.i18n import _


CLOUD_QUANTITY = 'CloudQuantity.'
CONVERTER = 'Converter.'
DESCRIPTIVE = 'Descriptive.'
PHENOMENON = 'Phenomenon.'
REMARK = 'Remark.'


def empty_if_none(code: str) -> str:
    return '' if code is None else code


class Command(abc.ABC):

    @abc.abstractmethod
    def execute(self, code: str, remark: list) -> tuple:
        pass

    @abc.abstractmethod
    def can_parse(self, code: str) -> any:
        pass


class CeilingHeightCommand(Command):
    regex = r'^CIG (\d{3})V(\d{3})\b'

    def __init__(self):
        self._pattern = re.compile(CeilingHeightCommand.regex)

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        min_ceiling = int(matches[0]) * 100
        max_ceiling = int(matches[1]) * 100
        remark.append(_('Remark.Ceiling.Height').format(min_ceiling, max_ceiling))
        return self._pattern.sub('', code, 1), remark


class CeilingSecondLocationCommand(Command):
    regex = r'^CIG (\d{3}) (\w+)\b'

    def __init__(self):
        self._pattern = re.compile(CeilingSecondLocationCommand.regex)

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        height = 100 * int(matches[0])
        remark.append(_('Remark.Ceiling.Second.Location').format(height, matches[1]))
        return self._pattern.sub('', code, 1), remark


class HailSizeCommand(Command):
    regex = r'^GR ((\d/\d)|((\d) ?(\d/\d)?))'

    def __init__(self):
        self._pattern = re.compile(HailSizeCommand.regex)

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Hail').format(matches[0]))
        return self._pattern.sub('', code, 1), remark


class HourlyMaximumMinimumTemperatureCommand(Command):
    regex = r'^4([01])(\d{3})([01])(\d{3})\b'

    def __init__(self):
        self._pattern = re.compile(HourlyMaximumMinimumTemperatureCommand.regex)

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Hourly.Maximum.Minimum.Temperature').format(
            convert_temperature_remarks(matches[0], matches[1]),
            convert_temperature_remarks(matches[2], matches[3])
        ))
        return self._pattern.sub('', code, 1), remark


class HourlyMaximumTemperatureCommand(Command):
    regex = r'^1([01])(\d{3})\b'

    def __init__(self):
        self._pattern = re.compile(HourlyMaximumTemperatureCommand.regex)

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Hourly.Maximum.Temperature').format(
            convert_temperature_remarks(matches[0], matches[1])
        ))
        return self._pattern.sub('', code, 1), remark


class HourlyMinimumTemperatureCommand(Command):
    regex = r'^2([01])(\d{3})\b'

    def __init__(self):
        self._pattern = re.compile(HourlyMinimumTemperatureCommand.regex)

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Hourly.Minimum.Temperature').format(
            convert_temperature_remarks(matches[0], matches[1])
        ))
        return self._pattern.sub('', code, 1), remark


class HourlyPrecipitationAmountCommand(Command):
    regex = r'^P(\d{4})\b'

    def __init__(self):
        self._pattern = re.compile(HourlyPrecipitationAmountCommand.regex)

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Precipitation.Amount.Hourly').format(
            int(matches[0])
        ))
        return self._pattern.sub('', code, 1), remark


class HourlyPressureCommand(Command):
    regex = r'^5(\d)(\d{3})\b'

    barometer_tendency = {
        0: 'Remark.Barometer.0',
        1: 'Remark.Barometer.1',
        2: 'Remark.Barometer.2',
        3: 'Remark.Barometer.3',
        4: 'Remark.Barometer.4',
        5: 'Remark.Barometer.5',
        6: 'Remark.Barometer.6',
        7: 'Remark.Barometer.7',
        8: 'Remark.Barometer.8'
    }

    def __init__(self):
        self._pattern = re.compile(HourlyPressureCommand.regex)

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(
            _(HourlyPressureCommand.barometer_tendency[int(matches[0])])
            + ' '
            + _('Remark.Pressure.Tendency').format(float(matches[1]) / 10)
        )
        return self._pattern.sub('', code, 1), remark


class HourlyTemperatureDewPointCommand(Command):
    regex = r'^T([01])(\d{3})(([01])(\d{3}))?'

    def __init__(self):
        self._pattern = re.compile(HourlyTemperatureDewPointCommand.regex)

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        if matches[2] is None:
            remark.append(_('Remark.Hourly.Temperature').format(convert_temperature_remarks(matches[0], matches[1])))
        else:
            remark.append(_('Remark.Hourly.Temperature.Dew.Point').format(
                convert_temperature_remarks(matches[0], matches[1]),
                convert_temperature_remarks(matches[3], matches[4])
            ))
        return self._pattern.sub('', code, 1), remark


class IceAccretionCommand(Command):
    regex = r'^l(\d)(\d{3})\b'

    def __init__(self):
        self._pattern = re.compile(IceAccretionCommand.regex)

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Ice.Accretion.Amount').format(int(matches[1]), int(matches[0])))
        return self._pattern.sub('', code, 1), remark


class ObscurationCommand(Command):
    regex = r'^([A-Z]{2}) ([A-Z]{3})(\d{3})'

    def __init__(self):
        self._pattern = re.compile(ObscurationCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        layer = _(CLOUD_QUANTITY + matches[1])
        height = 100 * int(matches[2])
        detail = _(PHENOMENON + matches[0])
        remark.append(_('Remark.Obscuration').format(layer, height, detail))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class PrecipitationAmount24HourCommand(Command):
    regex = r'^7(\d{4})\b'

    def __init__(self):
        self._pattern = re.compile(PrecipitationAmount24HourCommand.regex)

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Precipitation.Amount.24').format(convert_precipitation_amount(matches[0])))
        return self._pattern.sub('', code, 1), remark


class PrecipitationAmount36HourCommand(Command):
    regex = r'^([36])(\d{4})\b'

    def __init__(self):
        self._pattern = re.compile(PrecipitationAmount36HourCommand.regex)

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Precipitation.Amount.3.6').
                      format(matches[0], convert_precipitation_amount(matches[1])))
        return self._pattern.sub('', code, 1), remark


class PrecipitationBegCommand(Command):
    regex = r'^(([A-Z]{2})?([A-Z]{2})B(\d{2})?(\d{2}))'

    def __init__(self) -> None:
        super().__init__()
        self._pattern = re.compile(PrecipitationBegCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Precipitation.Beg').format(
            '' if matches[1] is None else _(DESCRIPTIVE + matches[1]),
            _(PHENOMENON + matches[2]),
            empty_if_none(matches[3]),
            matches[4]
        ))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class PrecipitationBegEndCommand(Command):
    regex = r'^(([A-Z]{2})?([A-Z]{2})B(\d{2})?(\d{2})E(\d{2})?(\d{2}))'

    def __init__(self):
        self._pattern = re.compile(PrecipitationBegEndCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Precipitation.Beg.End').format(
            '' if matches[1] is None else _(DESCRIPTIVE + matches[1]),
            _(PHENOMENON + matches[2]),
            empty_if_none(matches[3]),
            matches[4],
            empty_if_none(matches[5]),
            matches[6]
        ))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class PrecipitationEndCommand(Command):
    regex = r'^(([A-Z]{2})?([A-Z]{2})E(\d{2})?(\d{2}))'

    def __init__(self):
        self._pattern = re.compile(PrecipitationEndCommand.regex)

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Precipitation.End').format(
            '' if matches[1] is None else _(DESCRIPTIVE + matches[1]),
            _(PHENOMENON + matches[2]),
            empty_if_none(matches[3]),
            matches[4]
        ))
        return self._pattern.sub('', code, 1), remark


class PrevailingVisibilityCommand(Command):
    regex = r'^VIS ((\d)*( )?(\d?/?\d))V((\d)*( )?(\d?/?\d))'

    def __init__(self):
        self._pattern = re.compile(PrevailingVisibilityCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Variable.Prevailing.Visibility').format(
            matches[0], matches[4])
        )
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class SeaLevelPressureCommand(Command):
    regex = r'^SLP(\d{2})(\d)'

    def __init__(self):
        self._pattern = re.compile(SeaLevelPressureCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        pressure = '9' if matches[0].startswith('9') else '10'
        pressure += matches[0] + '.' + matches[1]
        remark.append(_('Remark.Sea.Level.Pressure').format(pressure))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class SecondLocationVisibilityCommand(Command):
    regex = r'^VIS ((\d)*( )?(\d?/?\d)) (\w+)'

    def __init__(self):
        self._pattern = re.compile(SecondLocationVisibilityCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Second.Location.Visibility').format(
            matches[0], matches[4])
        )
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class SectorVisibilityCommand(Command):
    regex = r'^VIS ([A-Z]{1,2}) ((\d)*( )?(\d?/?\d))'

    def __init__(self):
        self._pattern = re.compile(SectorVisibilityCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Sector.Visibility').format(
            _(CONVERTER + matches[0]),
            matches[1]
        ))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class SmallHailSizeCommand(Command):
    regex = r'^GR LESS THAN ((\d )?(\d/\d)?)'

    def __init__(self):
        self._pattern = re.compile(SmallHailSizeCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Hail.LesserThan').format(matches[0]))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class SnowDepthCommand(Command):
    regex = r'^4/(\d{3})'

    def __init__(self):
        self._pattern = re.compile(SnowDepthCommand.regex)

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Snow.Depth').format(int(matches[0])))
        return self._pattern.sub('', code, 1), remark


class SnowIncreaseCommand(Command):
    regex = r'^SNINCR (\d+)/(\d+)'

    def __init__(self):
        self._pattern = re.compile(SnowIncreaseCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Snow.Increasing.Rapidly').format(matches[0], matches[1]))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class SnowPelletsCommand(Command):
    regex = r'^GS (LGT|MOD|HVY)'

    def __init__(self):
        self._pattern = re.compile(SnowPelletsCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Snow.Pellets').format(_(REMARK + matches[0])))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class SunshineDurationCommand(Command):
    regex = r'^98(\d{3})'

    def __init__(self):
        self._pattern = re.compile(SunshineDurationCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Sunshine.Duration').format(int(matches[0])))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class SurfaceVisibilityCommand(Command):
    regex = r'^SFC VIS ((\d)*( )?(\d?/?\d))'

    def __init__(self):
        self._pattern = re.compile(SurfaceVisibilityCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Surface.Visibility').format(matches[0]))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class ThunderStormLocationCommand(Command):
    regex = r'^TS ([A-Z]{2})'

    def __init__(self):
        self._pattern = re.compile(ThunderStormLocationCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Thunderstorm.Location').format(_(CONVERTER + matches[0])))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class ThunderStormLocationMovingCommand(Command):
    regex = r'^TS ([A-Z]{2}) MOV ([A-Z]{2})'

    def __init__(self):
        self._pattern = re.compile(ThunderStormLocationMovingCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Thunderstorm.Location.Moving').format(
            _(CONVERTER + matches[0]), _(CONVERTER + matches[1])
        ))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class TornadicActivityBegCommand(Command):
    regex = '^(TORNADO|FUNNEL CLOUD|WATERSPOUT) (B(\d{2})?(\d{2}))( (\d+)? ([A-Z]{1,2})?)?'

    def __init__(self):
        self._pattern = re.compile(TornadicActivityBegCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Tornadic.Activity.Beginning').format(
            _(REMARK + matches[0].replace(' ', '')),
            empty_if_none(matches[2]),
            matches[3],
            matches[5],
            _(CONVERTER + matches[6])
        ))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class TornadicActivityBegEndCommand(Command):
    regex = r'^(TORNADO|FUNNEL CLOUD|WATERSPOUT) (B(\d{2})?(\d{2}))(E(\d{2})?(\d{2}))( (\d+)? ([A-Z]{1,2})?)?'

    def __init__(self):
        self._pattern = re.compile(TornadicActivityBegEndCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Tornadic.Activity.BegEnd').format(
            _(REMARK + matches[0].replace(' ', '')),
            empty_if_none(matches[2]),
            matches[3],
            empty_if_none(matches[5]),
            matches[6],
            matches[8],
            _(CONVERTER + matches[9])
        ))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class TornadicActivityEndCommand(Command):
    regex = r'^(TORNADO|FUNNEL CLOUD|WATERSPOUT) (E(\d{2})?(\d{2}))( (\d+)? ([A-Z]{1,2})?)?'

    def __init__(self):
        self._pattern = re.compile(TornadicActivityEndCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Tornadic.Activity.Ending').format(
            _(REMARK + matches[0].replace(' ', '')),
            empty_if_none(matches[2]),
            matches[3],
            matches[5],
            _(CONVERTER + matches[6])
        ))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class TowerVisibilityCommand(Command):
    regex = r'^TWR VIS ((\d)*( )?(\d?/?\d))'

    def __init__(self):
        self._pattern = re.compile(TowerVisibilityCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Tower.Visibility').format(matches[0]))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class VariableSkyCommand(Command):
    regex = r'^([A-Z]{3}) V ([A-Z]{3})'

    def __init__(self):
        self._pattern = re.compile(VariableSkyCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Variable.Sky.Condition').format(
            _(CLOUD_QUANTITY + matches[0]),
            _(CLOUD_QUANTITY + matches[1])
        ))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class VariableSkyHeightCommand(Command):
    regex = r'^([A-Z]{3})(\d{3}) V ([A-Z]{3})'

    def __init__(self):
        self._pattern = re.compile(VariableSkyHeightCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Variable.Sky.Condition.Height').format(
            100 * int(matches[1]),
            _(CLOUD_QUANTITY + matches[0]),
            _(CLOUD_QUANTITY + matches[2])
        ))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class VirgaDirectionCommand(Command):
    regex = r'^VIRGA ([A-Z]{2})'

    def __init__(self):
        self._pattern = re.compile(VirgaDirectionCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        match = self._pattern.search(code).groups()
        remark.append(_('Remark.Virga.Direction').format(_(CONVERTER + match[0])))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class WaterEquivalentSnowCommand(Command):
    regex = r'^933(\d{3})\b'

    def __init__(self):
        self._pattern = re.compile(WaterEquivalentSnowCommand.regex)

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.Water.Equivalent.Snow.Ground').format(
            float(matches[0]) / 10
        ))
        return self._pattern.sub('', code, 1), remark


class WindPeakCommand(Command):
    regex = r'^PK WND (\d{3})(\d{2,3})/(\d{2})?(\d{2})'

    def __init__(self):
        self._pattern = re.compile(WindPeakCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.PeakWind').format(
            matches[0],
            matches[1],
            empty_if_none(matches[2]),
            matches[3]
        ))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class WindShiftCommand(Command):
    regex = r'^WSHFT (\d{2})?(\d{2})'

    def __init__(self):
        self._pattern = re.compile(WindShiftCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.WindShift').format(
            empty_if_none(matches[0]),
            matches[1]
        ))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class WindShiftFropaCommand(Command):
    regex = r'^WSHFT (\d{2})?(\d{2}) FROPA'

    def __init__(self):
        self._pattern = re.compile(WindShiftFropaCommand.regex)

    def execute(self, code: str, remark: list) -> tuple:
        matches = self._pattern.search(code).groups()
        remark.append(_('Remark.WindShift.FROPA').format(
            empty_if_none(matches[0]),
            matches[1]
        ))
        return self._pattern.sub('', code, 1), remark

    def can_parse(self, code: str) -> any:
        return self._pattern.match(code)


class DefaultCommand(Command):
    def can_parse(self, code: str) -> any:
        return True

    def execute(self, code: str, remark: list) -> tuple:
        rmk_split = code.split(' ', 1)
        try:
            rem = _(REMARK + rmk_split[0])
            remark.append(rem)
        except TranslationError:
            remark.append(rmk_split[0])
        return '' if len(rmk_split) == 1 else rmk_split[1], remark


class RemarkCommandSupplier:

    def __init__(self):
        self.default_command = DefaultCommand()
        self._command_list = [WindPeakCommand(),
                              WindShiftFropaCommand(),
                              WindShiftCommand(),
                              TowerVisibilityCommand(),
                              SurfaceVisibilityCommand(),
                              PrevailingVisibilityCommand(),
                              SecondLocationVisibilityCommand(),
                              SectorVisibilityCommand(),
                              TornadicActivityBegEndCommand(),
                              TornadicActivityBegCommand(),
                              TornadicActivityEndCommand(),
                              PrecipitationBegEndCommand(),
                              PrecipitationBegCommand(),
                              PrecipitationEndCommand(),
                              ThunderStormLocationMovingCommand(),
                              ThunderStormLocationCommand(),
                              SmallHailSizeCommand(),
                              HailSizeCommand(),
                              SnowPelletsCommand(),
                              VirgaDirectionCommand(),
                              CeilingHeightCommand(),
                              ObscurationCommand(),
                              VariableSkyHeightCommand(),
                              VariableSkyCommand(),
                              CeilingSecondLocationCommand(),
                              SeaLevelPressureCommand(),
                              SnowIncreaseCommand(),
                              HourlyMaximumMinimumTemperatureCommand(),
                              HourlyMaximumTemperatureCommand(),
                              HourlyMinimumTemperatureCommand(),
                              HourlyPrecipitationAmountCommand(),
                              HourlyTemperatureDewPointCommand(),
                              HourlyPressureCommand(),
                              IceAccretionCommand(),
                              PrecipitationAmount36HourCommand(),
                              PrecipitationAmount24HourCommand(),
                              SnowDepthCommand(),
                              SunshineDurationCommand(),
                              WaterEquivalentSnowCommand()
                              ]

    def get(self, code: str) -> Command:
        for command in self._command_list:
            if command.can_parse(code):
                return command
        return self.default_command



import abc
from datetime import time

from metar_taf_parser.model.enum import Descriptive, Flag, WeatherChangeType, TimeIndicator, IcingIntensity, TurbulenceIntensity


class Country:

    def __init__(self, name):
        self._name = name

    def _get_name(self):
        """ Return the name of the country """
        return self._name

    def _set_name(self, value):
        """ Setter for the name """
        self._name = value

    name = property(_get_name, _set_name)


class Airport:
    def __init__(self):
        """ Initiate the attributes """
        self._name = None
        self._city = None
        self._country = None
        self._iata = None
        self._icao = None
        self._latitude = None
        self._longitude = None
        self._altitude = None
        self._timezone = None
        self._dst = None
        self._tz_database = None

    def _get_name(self):
        return self._name

    def _set_name(self, value):
        self._name = value

    def _get_city(self):
        return self._city

    def _set_city(self, value):
        self._city = value

    def _get_country(self):
        return self._country

    def _set_country(self, value):
        self._country = value

    def _get_iata(self):
        return self._iata

    def _set_iata(self, value):
        self._iata = value

    def _get_icao(self):
        return self._icao

    def _set_icao(self, value):
        self._icao = value

    def _get_latitude(self):
        return self._latitude

    def _set_latitude(self, value):
        self._latitude = value

    def _get_longitude(self):
        return self._longitude

    def _set_longitude(self, value):
        self._longitude = value

    def _set_altitude(self, value):
        self._altitude = value

    def _get_altitude(self):
        return self._altitude

    def _get_timezone(self):
        return self._timezone

    def _set_timezone(self, value):
        self._timezone = value

    def _get_dst(self):
        return self._dst

    def _set_dst(self, value):
        self._dst = value

    def _get_tz_database(self):
        return self._tz_database

    def _set_tz_database(self, value):
        self._tz_database = value

    name = property(_get_name, _set_name)
    city = property(_get_city, _set_city)
    country = property(_get_country, _set_country)
    iata = property(_get_iata, _set_iata)
    icao = property(_get_icao, _set_icao)
    latitude = property(_get_latitude, _set_latitude)
    longitude = property(_get_longitude, _set_longitude)
    altitude = property(_get_altitude, _set_altitude)
    timezone = property(_get_timezone, _set_timezone)
    dst = property(_get_dst, _set_dst)
    tz_database = property(_get_tz_database, _set_tz_database)


class Wind:
    def __init__(self):
        self._speed = None
        self._direction = None
        self._degrees = None
        self._gust = None
        self._min_variation = None
        self._max_variation = None
        self._unit = None

    def _get_speed(self):
        return self._speed

    def _set_speed(self, value):
        self._speed = value

    def _get_direction(self):
        return self._direction

    def _set_direction(self, value):
        self._direction = value

    def _get_degrees(self):
        return self._degrees

    def _set_degrees(self, value):
        self._degrees = value

    def _get_gust(self):
        return self._gust

    def _set_gust(self, value):
        self._gust = value

    def _get_min_variation(self):
        return self._min_variation

    def _set_min_variation(self, value):
        self._min_variation = value

    def _get_max_variation(self):
        return self._max_variation

    def _set_max_variation(self, value):
        self._max_variation = value

    def _get_unit(self):
        return self._unit

    def _set_unit(self, value):
        self._unit = value

    speed = property(_get_speed, _set_speed)
    direction = property(_get_direction, _set_direction)
    gust = property(_get_gust, _set_gust)
    degrees = property(_get_degrees, _set_degrees)
    unit = property(_get_unit, _set_unit)
    min_variation = property(_get_min_variation, _set_min_variation)
    max_variation = property(_get_max_variation, _set_max_variation)


class WindShear(Wind):
    def __init__(self):
        super().__init__()
        self._height = None

    def _get_height(self):
        return self._height

    def _set_height(self, value):
        self._height = value

    height = property(_get_height, _set_height)


class Visibility:
    def __init__(self):
        self._distance = None
        self._min_distance = None
        self._min_direction = None

    def _get_distance(self):
        return self._distance

    def _set_distance(self, value):
        self._distance = value

    def _get_min_distance(self):
        return self._min_distance

    def _set_min_distance(self, value):
        self._min_distance = value

    def _get_min_direction(self):
        return self._min_direction

    def _set_min_direction(self, value):
        self._min_direction = value

    distance = property(_get_distance, _set_distance)
    min_distance = property(_get_min_distance, _set_min_distance)
    min_direction = property(_get_min_direction, _set_min_direction)


class WeatherCondition:
    def __init__(self):
        self._intensity = None
        self._descriptive = None
        self._phenomenons = []

    def _get_intensity(self):
        return self._intensity

    def _set_intensity(self, value):
        self._intensity = value

    def _get_descriptive(self):
        return self._descriptive

    def _set_descriptive(self, value):
        self._descriptive = value

    def _get_phenomenons(self):
        return self._phenomenons

    def add_phenomenon(self, phenomenon):
        self._phenomenons.append(phenomenon)

    def is_valid(self):
        return len(self._phenomenons) != 0 or self._descriptive == Descriptive.THUNDERSTORM

    intensity = property(_get_intensity, _set_intensity)
    descriptive = property(_get_descriptive, _set_descriptive)
    phenomenons = property(_get_phenomenons)


class TemperatureDated:
    def __init__(self):
        self._temperature = None
        self._day = None
        self._hour = None

    def _get_temperature(self):
        return self._temperature

    def _set_temperature(self, value):
        self._temperature = value

    def _get_day(self):
        return self._day

    def _set_day(self, value):
        self._day = value

    def _get_hour(self):
        return self._hour

    def _set_hour(self, value):
        self._hour = value

    temperature = property(_get_temperature, _set_temperature)
    day = property(_get_day, _set_day)
    hour = property(_get_hour, _set_hour)


class RunwayInfo:

    def __init__(self):
        self._name = None
        self._min_range = None
        self._max_range = None
        self._trend = None

    def _get_name(self):
        return self._name

    def _set_name(self, value):
        self._name = value

    def _get_min_range(self):
        return self._min_range

    def _set_min_range(self, value):
        self._min_range = value

    def _get_max_range(self):
        return self._max_range

    def _set_max_range(self, value):
        self._max_range = value

    def _get_trend(self):
        return self._trend

    def _set_trend(self, value):
        self._trend = value

    name = property(_get_name, _set_name)
    min_range = property(_get_min_range, _set_min_range)
    max_range = property(_get_max_range, _set_max_range)
    trend = property(_get_trend, _set_trend)


class Cloud:

    def __init__(self):
        self._height = None
        self._quantity = None
        self._type = None

    def _get_height(self):
        return self._height

    def _set_height(self, value):
        self._height = value

    def _get_quantity(self):
        return self._quantity

    def _set_quantity(self, value):
        self._quantity = value

    def _set_type(self, value):
        self._type = value

    def _get_type(self):
        return self._type

    height = property(_get_height, _set_height)
    quantity = property(_get_quantity, _set_quantity)
    type = property(_get_type, _set_type)


class Icing:
    def __init__(self):
        self._intensity: IcingIntensity = None
        self._base_height = 0
        self._depth = 0

    def _get_intensity(self):
        return self._intensity

    def _set_intensity(self, intensity: IcingIntensity):
        self._intensity = intensity

    def _get_base_height(self):
        return self._base_height

    def _set_base_height(self, base_height: int):
        self._base_height = base_height

    def _get_depth(self):
        return self._depth

    def _set_depth(self, depth: int):
        self._depth = depth

    intensity = property(_get_intensity, _set_intensity)
    base_height = property(_get_base_height, _set_base_height)
    depth = property(_get_depth, _set_depth)


class Turbulence:

    def __init__(self):
        self._intensity: TurbulenceIntensity = None
        self._base_height = 0
        self._depth = 0

    def _get_intensity(self):
        return self._intensity

    def _set_intensity(self, intensity: TurbulenceIntensity):
        self._intensity = intensity

    def _get_base_height(self):
        return self._base_height

    def _set_base_height(self, base_height: int):
        self._base_height = base_height

    def _get_depth(self):
        return self._depth

    def _set_depth(self, depth: int):
        self._depth = depth

    intensity = property(_get_intensity, _set_intensity)
    base_height = property(_get_base_height, _set_base_height)
    depth = property(_get_depth, _set_depth)


class ITafGroups:
    def __init__(self):
        self._turbulence = []
        self._icings = []

    def _get_turbulence(self):
        return self._turbulence

    def _get_icings(self):
        return self._icings

    def add_turbulence(self, turbulence: Turbulence):
        self._turbulence.append(turbulence)

    def add_icing(self, icing: Icing):
        self._icings.append(icing)

    turbulence = property(_get_turbulence)
    icings = property(_get_icings)


class AbstractWeatherContainer(abc.ABC):

    def __init__(self):
        self._wind = None
        self._visibility = None
        self._vertical_visibility = None
        self._wind_shear = None
        self._cavok = None
        self._remark = None
        self._remarks = []
        self._clouds = []
        self._weather_conditions = []

    def _get_wind(self):
        return self._wind

    def _set_wind(self, value: Wind):
        self._wind = value

    def _get_visibility(self):
        return self._visibility

    def _set_visibility(self, value: Visibility):
        self._visibility = value

    def _get_vertical_visibility(self):
        return self._vertical_visibility

    def _set_vertical_visibility(self, value: int):
        self._vertical_visibility = value

    def _get_wind_shear(self):
        return self._wind_shear

    def _set_wind_shear(self, value: WindShear):
        self._wind_shear = value

    def _get_cavok(self):
        return self._cavok

    def _set_cavok(self, value: bool):
        self._cavok = value

    def _get_remark(self):
        return self._remark

    def _set_remark(self, value: str):
        self._remark = value

    def _get_remarks(self):
        return self._remarks

    def _set_remarks(self, remarks: list):
        self._remarks = remarks

    def _get_clouds(self):
        return self._clouds

    def add_cloud(self, cloud: Cloud):
        if cloud:
            self._clouds.append(cloud)

    def _get_weather_conditions(self):
        return self._weather_conditions

    def add_weather_condition(self, wc: WeatherCondition):
        if wc:
            self._weather_conditions.append(wc)
            return True

    wind = property(_get_wind, _set_wind)
    visibility = property(_get_visibility, _set_visibility)
    vertical_visibility = property(_get_vertical_visibility, _set_vertical_visibility)
    wind_shear = property(_get_wind_shear, _set_wind_shear)
    cavok = property(_get_cavok, _set_cavok)
    remark = property(_get_remark, _set_remark)
    remarks = property(_get_remarks, _set_remarks)
    clouds = property(_get_clouds)
    weather_conditions = property(_get_weather_conditions)


class AbstractValidity(abc.ABC):

    def _get_start_day(self):
        return self._start_day

    def _set_start_day(self, value: int):
        self._start_day = value

    def _get_start_hour(self):
        return self._start_hour

    def _set_start_hour(self, value: int):
        self._start_hour = value

    start_day = property(_get_start_day, _set_start_day)
    start_hour = property(_get_start_hour, _set_start_hour)


class AbstractWeatherCode(AbstractWeatherContainer):

    def __init__(self):
        super().__init__()
        self._day = None
        self._time = None
        self._airport = None
        self._message = None
        self._station = None
        self._flags = set()
        self._trends = []

    def _get_day(self):
        return self._day

    def _set_day(self, value: int):
        self._day = value

    def _get_time(self):
        return self._time

    def _set_time(self, value: time):
        self._time = value

    def _get_airport(self):
        return self._airport

    def _set_airport(self, value: Airport):
        self._airport = value

    def _get_message(self):
        return self._message

    def _set_message(self, value: str):
        self._message = value

    def _get_station(self):
        return self._station

    def _set_station(self, value: str):
        self._station = value

    def _get_trends(self):
        return self._trends

    def add_trend(self, value):
        self._trends.append(value)

    def _get_flags(self):
        return self._flags

    def _is_auto(self):
        return Flag.AUTO in self._flags

    def _is_amendment(self):
        return Flag.AMD in self._flags

    def _is_canceled(self):
        return Flag.CNL in self._flags

    def _is_corrected(self):
        return Flag.COR in self._flags

    def _is_nil(self):
        return Flag.NIL in self._flags

    day = property(_get_day, _set_day)
    time = property(_get_time, _set_time)
    airport = property(_get_airport, _set_airport)
    message = property(_get_message, _set_message)
    station = property(_get_station, _set_station)
    trends = property(_get_trends)
    flags = property(_get_flags)
    amendment = property(_is_amendment)
    auto = property(_is_auto)
    canceled = property(_is_canceled)
    corrected = property(_is_corrected)
    nil = property(_is_nil)


class Metar(AbstractWeatherCode):

    def __init__(self):
        super().__init__()
        self._temperature = None
        self._dew_point = None
        self._altimeter = None
        self._nosig = False
        self._runways_info = []

    def _get_temperature(self):
        return self._temperature

    def _set_temperature(self, input: int):
        self._temperature = input

    def _get_dew_point(self):
        return self._dew_point

    def _set_dew_point(self, value: int):
        self._dew_point = value

    def _get_altimeter(self):
        return self._altimeter

    def _set_altimeter(self, value: int):
        self._altimeter = value

    def _is_nosig(self):
        return self._nosig

    def _set_nosig(self, value: bool):
        self._nosig = value

    def _get_runways_info(self):
        return self._runways_info

    def add_runway_info(self, runway_info: RunwayInfo):
        self._runways_info.append(runway_info)

    temperature = property(_get_temperature, _set_temperature)
    dew_point = property(_get_dew_point, _set_dew_point)
    altimeter = property(_get_altimeter, _set_altimeter)
    nosig = property(_is_nosig, _set_nosig)
    runways_info = property(_get_runways_info)


class TAF(ITafGroups, AbstractWeatherCode):

    def __init__(self):
        ITafGroups.__init__(self)
        AbstractWeatherCode.__init__(self)
        self._validity = None
        self._max_temperature = None
        self._min_temperature = None

    def _get_validity(self):
        return self._validity

    def _set_validity(self, value: AbstractValidity):
        self._validity = value

    def _get_min_temperature(self):
        return self._min_temperature

    def _set_min_temperature(self, value: TemperatureDated):
        self._min_temperature = value

    def _get_max_temperature(self):
        return self._max_temperature

    def _set_max_temperature(self, value: TemperatureDated):
        self._max_temperature = value

    def becmgs(self):
        return list(filter(lambda trend: trend.type == WeatherChangeType.BECMG, self.trends))

    def probs(self):
        return list(filter(lambda trend: trend.type == WeatherChangeType.PROB, self.trends))

    def tempos(self):
        return list(filter(lambda trend: trend.type == WeatherChangeType.TEMPO, self.trends))

    def inters(self):
        return list(filter(lambda trend: trend.type == WeatherChangeType.INTER, self.trends))

    def fms(self):
        return list(filter(lambda trend: trend.type == WeatherChangeType.FM, self.trends))

    validity = property(_get_validity, _set_validity)
    max_temperature = property(_get_max_temperature, _set_max_temperature)
    min_temperature = property(_get_min_temperature, _set_min_temperature)


class AbstractTrend(AbstractWeatherContainer):
    def __init__(self, weather_change_type: WeatherChangeType):
        super().__init__()
        self._type = weather_change_type

    def _get_type(self):
        return self._type

    type = property(_get_type)


class MetarTrendTime:
    def __init__(self, time_indicator: TimeIndicator):
        self._type = time_indicator

    def _get_type(self):
        return self._type

    def _get_time(self):
        return self._time

    def _set_time(self, value: time):
        self._time = value

    type = property(_get_type)
    time = property(_get_time, _set_time)


class MetarTrend(AbstractTrend):

    def __init__(self, weather_change_type: WeatherChangeType):
        super().__init__(weather_change_type)
        self._times = []

    def _get_times(self):
        return self._times

    def add_time(self, value: MetarTrendTime):
        self._times.append(value)

    times = property(_get_times)


class TAFTrend(AbstractTrend, ITafGroups):
    def __init__(self, weather_change_type: WeatherChangeType):
        ITafGroups.__init__(self)
        AbstractTrend.__init__(self, weather_change_type)
        self._probability = None

    def _get_validity(self):
        return self._validity

    def _set_validity(self, value: AbstractValidity):
        self._validity = value

    def _get_probability(self):
        return self._probability

    def _set_probability(self, prob: int):
        self._probability = prob

    probability = property(_get_probability, _set_probability)
    validity = property(_get_validity, _set_validity)


class Validity(AbstractValidity):

    def __init__(self):
        super().__init__()
        self._end_hour = None

    def _get_end_hour(self):
        return self._end_hour

    def _set_end_hour(self, value: int):
        self._end_hour = value

    def _get_end_day(self):
        return self._end_day

    def _set_end_day(self, value: int):
        self._end_day = value

    end_hour = property(_get_end_hour, _set_end_hour)
    end_day = property(_get_end_day, _set_end_day)


class FMValidity(AbstractValidity):

    def __init__(self):
        super().__init__()
        self._start_minutes = None

    def _get_start_minutes(self):
        return self._start_minutes

    def _set_start_minutes(self, value: int):
        self._start_minutes = value

    start_minutes = property(_get_start_minutes, _set_start_minutes)



def degrees_to_cardinal(input):
    if input.isnumeric():
        degrees = int(input)
        dirs = ["N", "NNE", "NE", "ENE", "E", "ESE", "SE", "SSE",
                "S", "SSW", "SW", "WSW", "W", "WNW", "NW", "NNW"]
        ix = int((degrees + 11.25) / 22.5)
        return dirs[ix % 16]
    else:
        return 'VRB'


def convert_visibility(input):
    if '9999' == input:
        return '> 10km'
    return str(int(input)) + 'm'


def convert_temperature(input):
    if input.startswith('M'):
        return -int(input.split('M')[1])
    return int(input)


def convert_inches_mercury_to_pascal(input):
    return 33.8639 * input


def convert_temperature_remarks(sign: str, temperature: str):
    temp = float(temperature) / 10
    return temp if '0' == sign else -1 * temp


def convert_precipitation_amount(amount: str):
    return float(amount) / 100



import re

from metar_taf_parser.model.enum import IcingIntensity, TurbulenceIntensity
from metar_taf_parser.model.model import ITafGroups, Icing, Turbulence


class IcingCommand:
    regex = r'^6(\d)(\d{3})(\d)$'

    def __init__(self):
        self._pattern = re.compile(IcingCommand.regex)

    def can_parse(self, input: str):
        return self._pattern.search(input)

    def execute(self, itaf: ITafGroups, input: str):
        """

        :param itaf: metar_taf_parser.model.model.ITafGroups
        :param input: string
        :return:
        """
        matches = self._pattern.search(input).groups()
        icing = Icing()
        icing.intensity = IcingIntensity(matches[0])
        icing.base_height = 100 * int(matches[1])
        icing.depth = 1000 * int(matches[2])
        itaf.add_icing(icing)


class TurbulenceCommand:
    regex = r"^5(\d|'X')(\d{3})(\d)$"

    def __init__(self):
        self._pattern = re.compile(TurbulenceCommand.regex)

    def can_parse(self, input: str):
        return self._pattern.search(input)

    def execute(self, itaf: ITafGroups, input: str):
        """

        :param itaf: metar_taf_parser.model.model.ITafGroups
        :param input: string
        :return:
        """
        matches = self._pattern.search(input).groups()
        turbulence = Turbulence()
        turbulence.intensity = TurbulenceIntensity(matches[0])
        turbulence.base_height = 100 * int(matches[1])
        turbulence.depth = 1000 * int(matches[2])
        itaf.add_turbulence(turbulence)


class TAFCommandSupplier:
    def __init__(self):
        self._commands = [IcingCommand(), TurbulenceCommand()]

    def get(self, input: str):
        for command in self._commands:
            if command.can_parse(input):
                return command




import re

from metar_taf_parser.commons import converter
from metar_taf_parser.commons.converter import convert_visibility
from metar_taf_parser.model.enum import CloudQuantity, CloudType
from metar_taf_parser.model.model import Visibility, Wind, WindShear, Cloud, AbstractWeatherContainer


def set_wind_elements(wind: Wind, direction: str, speed: str, gust: str, unit: str):
    """
    This function updates a wind element.
    :param wind: Wind. The wind object
    :param direction: str. The direction in degrees
    :param speed: str. The speed
    :param gust: int. The speed of the gust.
    :param unit: str. The speed unit
    :return: None.
    """
    wind.speed = int(speed)
    wind.direction = converter.degrees_to_cardinal(direction)

    if 'VRB' != direction:
        wind.degrees = int(direction)
    if gust:
        wind.gust = int(gust)
    if unit:
        wind.unit = unit
    else:
        wind.unit = 'KT'


class CloudCommand:
    cloud_regex = r'^([A-Z]{3})((\d{3}|/{3})([A-Z]{2,3}|/{3})?)?$'
    undefined = '///'

    def __init__(self):
        self._pattern = re.compile(CloudCommand.cloud_regex)

    def parse(self, cloud_string: str):
        m = self._pattern.search(cloud_string).groups()
        cloud = Cloud()
        try:
            if CloudQuantity[m[0]]:
                cloud.quantity = CloudQuantity[m[0]]
            if m[2] and m[2] != CloudCommand.undefined:
                cloud.height = 100 * int(m[2])
            if m[3] and m[3] != CloudCommand.undefined:
                cloud.type = CloudType[m[3]]
            return cloud
        except KeyError:
            return

    def execute(self, container: AbstractWeatherContainer, cloud_string: str):
        cloud = self.parse(cloud_string)
        if cloud and cloud.quantity:
            container.add_cloud(cloud)
            return True

    def can_parse(self, cloud_string: str):
        return self._pattern.search(cloud_string)


class MainVisibilityCommand:
    regex = r'^(\d{4})(|NDV)$'

    def __init__(self):
        self._pattern = re.compile(MainVisibilityCommand.regex)

    def can_parse(self, visibility_string: str):
        return self._pattern.search(visibility_string)

    def execute(self, container: AbstractWeatherContainer, visibility_string: str):
        matches = self._pattern.search(visibility_string).groups()
        if container.visibility is None:
            container.visibility = Visibility()
        container.visibility.distance = convert_visibility(matches[0])
        return True


class WindCommand:
    regex = r'^(VRB|[0-3]\d{2})(\d{2})G?(\d{2})?(KT|MPS|KM\/H)?'

    def __init__(self):
        self._pattern = re.compile(WindCommand.regex)

    def can_parse(self, wind_string: str):
        """

        :param wind_string: str
            The string to parse
        :return:
        """
        return self._pattern.search(wind_string)

    def parse_wind(self, wind_string: str):
        wind = Wind()
        matches = self._pattern.search(wind_string).groups()
        set_wind_elements(wind, matches[0], matches[1], matches[2], matches[3])
        return wind

    def execute(self, container: AbstractWeatherContainer, wind_string: str):
        wind = self.parse_wind(wind_string)
        container.wind = wind
        return True


class WindVariationCommand:
    regex = r'^(\d{3})V(\d{3})'

    def __init__(self):
        self._pattern = re.compile(WindVariationCommand.regex)

    def can_parse(self, wind_string: str):
        return self._pattern.search(wind_string)

    def parse_wind_variation(self, wind: Wind, wind_string: str):
        matches = self._pattern.search(wind_string).groups()
        wind.min_variation = int(matches[0])
        wind.max_variation = int(matches[1])

    def execute(self, container, wind_string):
        self.parse_wind_variation(container.wind, wind_string)
        return True


class WindShearCommand:
    regex = r'^WS(\d{3})\/(\w{3})(\d{2})G?(\d{2})?(KT|MPS|KM\/H)'

    def __init__(self):
        self._pattern = re.compile(WindShearCommand.regex)

    def can_parse(self, wind_string: str):
        return self._pattern.search(wind_string)

    def parse_wind_shear(self, wind_string: str):
        wind_shear = WindShear()
        matches = self._pattern.search(wind_string).groups()

        wind_shear.height = 100 * int(matches[0])
        set_wind_elements(wind_shear, matches[1], matches[2], matches[3], matches[4])
        return wind_shear

    def execute(self, container: AbstractWeatherContainer, wind_string: str):
        container.wind_shear = self.parse_wind_shear(wind_string)
        return True


class VerticalVisibilityCommand:

    regex = r'^VV(\d{3})$'

    def __init__(self):
        self._pattern = re.compile(VerticalVisibilityCommand.regex)

    def execute(self, container: AbstractWeatherContainer, visibility_string: str):
        matches = self._pattern.search(visibility_string).groups()
        container.vertical_visibility = 100 * int(matches[0])
        return True

    def can_parse(self, visibility_string: str):
        return self._pattern.search(visibility_string)


class MinimalVisibilityCommand:
    regex = r'^(\d{4}[a-z])$'

    def __init__(self):
        self._pattern = re.compile(MinimalVisibilityCommand.regex)

    def can_parse(self, visibility_string: str):
        return self._pattern.search(visibility_string)

    def execute(self, container: AbstractWeatherContainer, visibility_string: str):
        """

        :param container: AbstractWeatherContainer
        :param visibility_string: string
        :return:
        """
        matches = self._pattern.search(visibility_string).groups()
        container.visibility.min_distance = int(matches[0][0:4])
        container.visibility.min_direction = matches[0][4]
        return True


class MainVisibilityNauticalMilesCommand:

    regex = r'^(P|M)?(\d)*(\s)?((\d\/\d)?SM)$'

    def __init__(self):
        self._pattern = re.compile(MainVisibilityNauticalMilesCommand.regex)

    def can_parse(self, wind_string: str):
        return self._pattern.search(wind_string)

    def execute(self, container: AbstractWeatherContainer, visibility_string: str):
        if container.visibility is None:
            container.visibility = Visibility()
        container.visibility.distance = visibility_string
        return True


class CommandSupplier:

    def __init__(self):
        self._commands = [
            WindShearCommand(), WindCommand(), WindVariationCommand(), MainVisibilityCommand(),
            MainVisibilityNauticalMilesCommand(), MinimalVisibilityCommand(),
            VerticalVisibilityCommand(), CloudCommand()
        ]

    def get(self, input: str):
        for command in self._commands:
            if command.can_parse(input):
                return command



import abc
import re
from datetime import time

from metar_taf_parser.command.common import CommandSupplier
from metar_taf_parser.command.metar import CommandSupplier as MetarCommandSupplier
from metar_taf_parser.command.remark import RemarkCommandSupplier
from metar_taf_parser.command.taf import TAFCommandSupplier
from metar_taf_parser.commons import converter
from metar_taf_parser.commons.exception import TranslationError
from metar_taf_parser.model.enum import Flag, Intensity, Descriptive, Phenomenon, TimeIndicator, WeatherChangeType
from metar_taf_parser.model.model import WeatherCondition, Visibility, Metar, TemperatureDated, \
    AbstractWeatherContainer, TAF, TAFTrend, MetarTrend, Validity, FMValidity, MetarTrendTime


def parse_delivery_time(abstract_weather_code, time_string):
    """
    Parses the delivery time of a METAR/TAF
    :param abstract_weather_code: The TAF or METAR object
    :param time_string: The string representing the delivery time
    :return: None
    """
    abstract_weather_code.day = int(time_string[0:2])
    abstract_weather_code.time = time(int(time_string[2:4]), int(time_string[4:6]))


def _parse_flags(abstract_weather_code, flag_string):
    try:
        abstract_weather_code.flags.add(Flag(flag_string))
        return True
    except ValueError:
        return False


def parse_remark(container: AbstractWeatherContainer, line: list, index: int):
    """
    This function parses the array containing the remark and concat the array into a string
    :param container: the metar, taf or taf trend to update
    :param line: The array containing the current line tokens
    :param index: the index starting the remark ie token RMK
    :return: None
    """
    remarks = RemarkParser().parse(str.join(' ', line[index + 1:]))
    container.remarks = remarks
    container.remark = str.join(' ', remarks)


def _parse_temperature(input: str):
    """
    Parses the temperature in a TAF
    :param input: the string containing the temperature
    :return: TemperatureDated object
    """
    parts = input.split('/')
    temperature = TemperatureDated()

    temperature.temperature = converter.convert_temperature(parts[0][2:])
    temperature.day = int(parts[1][0:2])
    temperature.hour = int(parts[1][2:4])
    return temperature


def _parse_validity(input: str):
    """
    Parses validity of a TAF or a TAFTrend
    :param input: the string containing the validity
    :return: Validity object
    """
    validity = Validity()
    parts = input.split('/')
    validity.start_day = int(parts[0][0:2])
    validity.start_hour = int(parts[0][2:])
    validity.end_day = int(parts[1][0:2])
    validity.end_hour = int(parts[1][2:])
    return validity


def _parse_from_validity(input: str):
    """
    Parses the validity for a FROM taf trend
    :param input: the string containing the validity
    :return: a Validity object
    """
    validity = FMValidity()
    validity.start_day = int(input[2:4])
    validity.start_hour = int(input[4:6])
    validity.start_minutes = int(input[6:8])

    return validity


class AbstractParser(abc.ABC):
    """
    Abstract class.
    Base parser.
    """
    FM = 'FM'
    TEMPO = 'TEMPO'
    BECMG = 'BECMG'
    RMK = 'RMK'
    TOKENIZE_REGEX = r'\s((?=\d\/\dSM)(?<!\s\d\s)|(?!\d\/\dSM))|='
    INTENSITY_REGEX = r'^(-|\+|VC)'
    CAVOK = 'CAVOK'

    def __init__(self):
        self._common_supplier = CommandSupplier()
        self._tokenize_regex_pattern = re.compile(AbstractParser.TOKENIZE_REGEX)
        self._intensity_regex_pattern = re.compile(AbstractParser.INTENSITY_REGEX)

    @abc.abstractmethod
    def parse(self, input: str):
        pass

    def _parse_weather_condition(self, input: str):
        """
        Parses a string into a weather condition. The result is not necessarily valid
        :param input: The input to parse
        :return: WeatherCondition object
        """
        weather_condition = WeatherCondition()
        if self._intensity_regex_pattern.match(input):
            match = self._intensity_regex_pattern.findall(input)[0]
            weather_condition.intensity = Intensity(match)
            input = input[len(match):]

        for name, member in Descriptive.__members__.items():
            if member.value in input:
                weather_condition.descriptive = member
                input = input[len(member.value):]

        previous_token = ''
        while input != '' and input != previous_token:
            previous_token = input
            for name, member in Phenomenon.__members__.items():
                if re.match(r'^' + member.value, input):
                    weather_condition.add_phenomenon(member)
                    input = input[len(member.value):]

        return weather_condition if input == '' and weather_condition.is_valid() else None

    def tokenize(self, input: str):
        """
        Parses the message into different tokens
        :param input: The metar or TAF as string
        :return: List of tokens
        """
        return list(filter(None, self._tokenize_regex_pattern.split(input)))

    def general_parse(self, abstract_weather_container: AbstractWeatherContainer, input: str):
        """
        Common parse method for METAR, TAF and trends object
        :param abstract_weather_container: the object to update
        :param input: The token to parse
        :return: True if the token was parsed false otherwise
        """
        if AbstractParser.CAVOK == input:
            abstract_weather_container.cavok = True
            if abstract_weather_container.visibility is None:
                abstract_weather_container.visibility = Visibility()
            abstract_weather_container.visibility.distance = '> 10km'
            return True

        command = self._common_supplier.get(input)
        if command:
            return command.execute(abstract_weather_container, input)

        return abstract_weather_container.add_weather_condition(self._parse_weather_condition(input))


class MetarParser(AbstractParser):
    """
    Parser to Metar messages.
    """
    AT = 'AT'
    TL = 'TL'

    def __init__(self):
        super().__init__()
        self._metar_command_supplier = MetarCommandSupplier()

    def _parse_trend(self, index: int, trend: MetarTrend, trend_parts: list):
        """
        Parses a trend of a metar
        :param index: the index starting the trend in the list
        :param trend: The trend to update
        :param trend_parts: string[] array of tokens
        :return: the last index of the token that was last parsed
        """
        i = index + 1
        while i < len(trend_parts) and AbstractParser.TEMPO != trend_parts[i] and AbstractParser.BECMG != trend_parts[i]:
            if trend_parts[i].startswith(AbstractParser.FM) or trend_parts[i].startswith(MetarParser.TL) or trend_parts[i].startswith(MetarParser.AT):

                trend_time = MetarTrendTime(TimeIndicator[trend_parts[i][0:2]])
                trend_time.time = time(int(trend_parts[i][2:4]), int(trend_parts[i][4:6]))
                trend.add_time(trend_time)
            else:
                self.general_parse(trend, trend_parts[i])
            i = i + 1
        return i - 1

    def parse(self, input: str):
        """
        Parses an message and returns a METAR
        :param input: The message to parse
        :return: METAR
        """
        metar = Metar()

        metar_tab = self.tokenize(input)
        metar.station = metar_tab[0]

        metar.message = input

        parse_delivery_time(metar, metar_tab[1])
        index = 2
        while index < len(metar_tab):
            if not super().general_parse(metar, metar_tab[index]) and not _parse_flags(metar, metar_tab[index]):
                if 'NOSIG' == metar_tab[index]:
                    metar.nosig = True
                elif AbstractParser.TEMPO == metar_tab[index] or AbstractParser.BECMG == metar_tab[index]:
                    trend = MetarTrend(WeatherChangeType[metar_tab[index]])
                    index = self._parse_trend(index, trend, metar_tab)
                    metar.add_trend(trend)
                elif AbstractParser.RMK == metar_tab[index]:
                    parse_remark(metar, metar_tab, index)
                    break
                else:
                    command = self._metar_command_supplier.get(metar_tab[index])
                    if command:
                        command.execute(metar, metar_tab[index])
            index = index + 1
        return metar


class TAFParser(AbstractParser):
    """
    Parser of TAF messages.
    """
    TAF = 'TAF'
    PROB = 'PROB'
    TX = 'TX'
    TN = 'TN'

    def __init__(self):
        super().__init__()
        self._validity_pattern = re.compile(r'^\d{4}/\d{4}$')
        self._taf_command_supplier = TAFCommandSupplier()

    def parse(self, input: str):
        """
        Parses a message into a TAF
        :param input: the message to parse
        :return: a TAF object or None if the message is invalid
        """
        taf = TAF()
        lines = self._extract_lines_tokens(input)
        if TAFParser.TAF != lines[0][0]:
            return
        index = 1
        if TAFParser.TAF == lines[0][1]:
            index = 2
        if _parse_flags(taf, lines[0][index]):
            index += 1

        taf.station = lines[0][index]
        index += 1
        taf.message = input
        parse_delivery_time(taf, lines[0][index])
        index += 1
        taf.validity = _parse_validity(lines[0][index])

        for i in range(index + 1, len(lines[0])):
            token = lines[0][i]
            command = self._taf_command_supplier.get(token)
            if AbstractParser.RMK == token:
                parse_remark(taf, lines[0], i)
                break
            elif token.startswith(TAFParser.TX):
                taf.max_temperature = _parse_temperature(token)
            elif token.startswith(TAFParser.TN):
                taf.min_temperature = _parse_temperature(token)
            elif command:
                command.execute(taf, token)
            else:
                _parse_flags(taf, token)
                self.general_parse(taf, token)

        # Handle the other lines
        for line in lines[1:]:
            self._parse_line(taf, line)

        return taf

    def _extract_lines_tokens(self, taf_code: str):
        """
        Format the message as a multiple line code so each line can be parsed
        :param taf_code: The base message
        :return: a list of string representing the lines of the message.
        """
        single_line = taf_code.replace('\n', ' ')
        clean_line = re.sub(r'\s{2,}', ' ', single_line)
        lines = re.sub(r'\s(PROB\d{2}\sTEMPO|TEMPO|INTER|BECMG|FM|PROB)', '\n\g<1>', clean_line).splitlines()
        lines_token = [self.tokenize(line) for line in lines]

        if len(lines_token) > 1:
            last_line = lines_token[len(lines) - 1]
            temperatures = list(filter(lambda x: x.startswith(TAFParser.TX) or x.startswith(TAFParser.TN), last_line))

            if temperatures:
                lines_token[0] = lines_token[0] + temperatures
                lines_token[len(lines) - 1] = list(filter(lambda x: not x.startswith(TAFParser.TX) and not x.startswith(TAFParser.TN), last_line))
        return lines_token

    def _parse_line(self, taf: TAF, line_tokens: list):
        """
        Parses the tokens of the line and updates the TAF object.
        :param taf: TAF object to update
        :param line_tokens: the array of tokens representing a line
        :return: None
        """
        index = 1
        if line_tokens[0].startswith(TAFParser.FM):
            trend = TAFTrend(WeatherChangeType.FM)
            trend.validity = _parse_from_validity(line_tokens[0])
        elif line_tokens[0].startswith(TAFParser.PROB):
            trend = TAFTrend(WeatherChangeType.PROB)
            if len(line_tokens) > 1 and TAFParser.TEMPO == line_tokens[1]:
                trend = TAFTrend(WeatherChangeType(line_tokens[1]))
                index = 2
            trend.probability = int(line_tokens[0][4:])
        else:
            trend = TAFTrend(WeatherChangeType(line_tokens[0]))
        self._parse_trend(index, line_tokens, trend)
        taf.add_trend(trend)

    def _parse_trend(self, index: int, line: list, trend: TAFTrend):
        """
        Parses a trend of the TAF
        :param index: the index at which the array should be parsed
        :param line: The array of string containing the line
        :param trend: The trend object to update
        :return: None
        """
        for i in range(index, len(line)):
            command = self._taf_command_supplier.get(line[i])

            if command:
                command.execute(trend, line[i])
            elif AbstractParser.RMK == line[i]:
                parse_remark(trend, line, i)
                break
            elif self._validity_pattern.search(line[i]):
                trend.validity = _parse_validity(line[i])
            else:
                super().general_parse(trend, line[i])


class RemarkParser:
    def __init__(self):
        self._supplier = RemarkCommandSupplier()

    def parse(self, code: str) -> list:
        rmk_str = code
        rmk_list = []

        while rmk_str:
            try:
                (rmk_str, rmk_list) = self._supplier.get(rmk_str).execute(rmk_str, rmk_list)
            except TranslationError:
                (rmk_str, rmk_list) = self._supplier.default_command.execute(rmk_str, rmk_list)
        return rmk_list



import gettext
import locale
import os

from metar_taf_parser.commons.exception import TranslationError


localedir = os.path.join(os.path.abspath(os.path.dirname(__file__)), '../locale')
langAvailable = os.listdir(localedir)
lang = locale.getdefaultlocale()[0][:2]
if lang not in langAvailable:
    lang = 'en'
t = gettext.translation(domain='messages', localedir=localedir, fallback=True, languages=[lang])


def _(message: str) -> str:
    translation = t.gettext(message)
    if message == translation:
        raise TranslationError(translation=message, message='Missing translation')
    return translation



class TranslationError(Exception):
    """Exception raised when a translation is not available
    Attributes:
        translation -- the missing string in the translation file
        message -- explanation of the error
    """
    def __init__(self, translation: str, message: str):
        self.message = message
        self.translation = translation



import enum

from metar_taf_parser.commons.i18n import _


class CloudQuantity(enum.Enum):
    SKC = 'SKC'   # Sky clear
    FEW = 'FEW'   # Few
    BKN = 'BKN'  # Broken
    SCT = 'SCT'  # Scattered
    OVC = 'OVC'  # Overcast
    NSC = 'NSC'  # No significant cloud

    def __str__(self):
        return _('CloudQuantity.' + self.value)


class CloudType(enum.Enum):
    CB = 'CB'  # Cumulonimbus
    TCU = 'TCU'  # Towering cumulus, cumulus congestus
    CI = 'CI'  # Cirrus
    CC = 'CC'  # Cirrocumulus
    CS = 'CS'  # Cirrostratus
    AC = 'AC'  # Altocumulus
    ST = 'ST'  # Stratus
    CU = 'CU'  # Cumulus
    AS = 'AS'  # Astrostratus
    NS = 'NS'  # Nimbostratus
    SC = 'SC'  # Stratocumulus

    def __str__(self):
        return _('CloudType.' + self.value)


class Flag(enum.Enum):
    AMD = 'AMD'
    AUTO = 'AUTO'
    CNL = 'CNL'
    COR = 'COR'
    NIL = 'NIL'

    def __str__(self):
        return _('Flag.' + self.value)


class IcingIntensity(enum.Enum):
    NONE = '0'
    LIGHT = '1'
    LIGHT_RIME_ICING_CLOUD = '2'
    LIGHT_CLEAR_ICING_PRECIPITATION = '3'
    MODERATE_MIXED_ICING = '4'
    MODERATE_RIME_ICING_CLOUD = '5'
    MODERATE_CLEAR_ICING_PRECIPITATION = '6'
    SEVERE_MIXED_ICING = '7'
    SEVERE_RIME_ICING_CLOUD = '8'
    SEVERE_CLEAR_ICING_PRECIPITATION = '9'

    def __str__(self) -> str:
        return _('IcingIntensity.' + self.value)


class Intensity(enum.Enum):
    LIGHT = '-'
    HEAVY = '+'
    IN_VICINITY = 'VC'

    def __str__(self):
        return _('Intensity.' + self.value)


class Descriptive(enum.Enum):
    SHOWERS = 'SH'
    SHALLOW = 'MI'
    PATCHES = 'BC'
    PARTIAL = 'PR'
    DRIFTING = 'DR'
    THUNDERSTORM = 'TS'
    BLOWING = 'BL'
    FREEZING = 'FZ'

    def __str__(self):
        return _('Descriptive.' + self.value)


class Phenomenon(enum.Enum):
    RAIN = 'RA'
    DRIZZLE = 'DZ'
    SNOW = 'SN'
    SNOW_GRAINS = 'SG'
    ICE_PELLETS = 'PL'
    ICE_CRYSTALS = 'IC'
    HAIL = 'GR'
    SMALL_HAIL = 'GS'
    UNKNOW_PRECIPITATION = 'UP'
    FOG = 'FG'
    VOLCANIC_ASH = 'VA'
    MIST = 'BR'
    HAZE = 'HZ'
    WIDESPREAD_DUST = 'DU'
    SMOKE = 'FU'
    SAND = 'SA'
    SPRAY = 'PY'
    SQUALL = 'SQ'
    SAND_WHIRLS = 'PO'
    DUSTSTORM = 'DS'
    SANDSTORM = 'SS'
    FUNNEL_CLOUD = 'FC'

    def __str__(self):
        return _('Phenomenon.' + self.value)


class TimeIndicator(enum.Enum):
    AT = 'AT'
    FM = 'FM'
    TL = 'TL'

    def __str__(self):
        return _('TimeIndicator.' + self.value)


class TurbulenceIntensity(enum.Enum):
    NONE = '0'
    LIGHT = '1'
    MODERATE_CLEAR_AIR_OCCASIONAL = '2'
    MODERATE_CLEAR_AIR_FREQUENT = '3'
    MODERATE_CLOUD_OCCASIONAL = '4'
    MODERATE_CLOUD_FREQUENT = '5'
    SEVERE_CLEAR_AIR_OCCASIONAL = '6'
    SEVERE_CLEAR_AIR_FREQUENT = '7'
    SEVERE_CLOUD_OCCASIONAL = '8'
    SEVERE_CLOUD_FREQUENT = '9'
    EXTREME = 'X'

    def __str__(self) -> str:
        return _('TurbuleneIntensity.' + self.value)


class WeatherChangeType(enum.Enum):
    FM = 'FM'
    BECMG = 'BECMG'
    TEMPO = 'TEMPO'
    PROB = 'PROB'
    INTER = 'INTER'

    def __str__(self):
        return _('WeatherChangeType.' + self.value)



import re

from metar_taf_parser.commons import converter
from metar_taf_parser.model.model import RunwayInfo, Metar


class AltimeterCommand:
    regex = r'^Q(\d{4})$'

    def __init__(self):
        self._pattern = re.compile(AltimeterCommand.regex)

    def can_parse(self, input: str):
        return self._pattern.search(input)

    def execute(self, metar: Metar, input: str):
        """

        :param metar: metar_taf_parser.model.model.Metar
        :param input: string
        :return:
        """
        matches = self._pattern.search(input).groups()
        metar.altimeter = int(matches[0])


class AltimeterMercuryCommand:
    regex = r'^A(\d{4})$'

    def __init__(self):
        self._pattern = re.compile(AltimeterMercuryCommand.regex)

    def can_parse(self, input: str):
        return self._pattern.search(input)

    def execute(self, metar: Metar, input: str):
        matches = self._pattern.search(input).groups()
        mercury = float(matches[0]) / 100
        metar.altimeter = int(converter.convert_inches_mercury_to_pascal(mercury))


class RunwayCommand:
    generic_regex = r'^(R\d{2}\w?/)'
    runway_max_range_regex = r'^R(\d{2}\w?)/(\d{4})V(\d{3})(\w{0,2})'
    runway_regex = r'^R(\d{2}\w?)/(\w)?(\d{4})(\w{0,2})$'

    def __init__(self):
        self._generic_pattern = re.compile(RunwayCommand.generic_regex)
        self._max_range_pattern = re.compile(RunwayCommand.runway_max_range_regex)
        self._runway_pattern = re.compile(RunwayCommand.runway_regex)

    def can_parse(self, input: str):
        return self._generic_pattern.match(input)

    def execute(self, metar: Metar, input: str):
        matches = self._runway_pattern.findall(input)
        runway = RunwayInfo()
        if matches:
            runway.name = matches[0][0]
            runway.min_range = int(matches[0][2])
            runway.trend = matches[0][3]
            metar.add_runway_info(runway)

        matches = self._max_range_pattern.findall(input)
        if matches:
            runway.name = matches[0][0]
            runway.min_range = int(matches[0][1])
            runway.max_range = int(matches[0][2])
            runway.trend = matches[0][3]
            metar.add_runway_info(runway)


class TemperatureCommand:
    regex = r'^(M?\d{2})/(M?\d{2})$'

    def __init__(self):
        self._pattern = re.compile(TemperatureCommand.regex)

    def can_parse(self, input: str):
        return self._pattern.match(input)

    def execute(self, metar: Metar, input: str):
        matches = self._pattern.search(input).groups()
        metar.temperature = converter.convert_temperature(matches[0])
        metar.dew_point = converter.convert_temperature(matches[1])


class CommandSupplier:
    def __init__(self):
        self._commands = [RunwayCommand(), TemperatureCommand(), AltimeterCommand(), AltimeterMercuryCommand()]

    def get(self, input: str):
        for command in self._commands:
            if command.can_parse(input):
                return command
